from .forward_pass import validate_network
from .generic import create_pytorch_objects
import os, sys
from typing import Optional
from pathlib import Path
import pandas as pd

# hides torchio citation request, see https://github.com/fepegar/torchio/issues/235
os.environ["TORCHIO_HIDE_CITATION_PROMPT"] = "1"

import torch
import cv2
import numpy as np
from torch.utils.data import DataLoader
from skimage.io import imsave
from tqdm import tqdm
from torch.cuda.amp import autocast
import openslide
from GANDLF.data import get_testing_loader
from GANDLF.utils import (
    BEST_MODEL_PATH_END,
    LATEST_MODEL_PATH_END,
    load_ov_model,
    print_model_summary,
    applyCustomColorMap,
)

from GANDLF.data.inference_dataloader_histopath import InferTumorSegDataset
from GANDLF.data.preprocessing import get_transforms_for_preprocessing


def inference_loop(
    inferenceDataFromPickle: pd.DataFrame,
    device: str,
    parameters: dict,
    modelDir: str,
    outputDir: Optional[str] = None,
) -> None:
    """
    The main training loop.

    Args:
        inferenceDataFromPickle (pandas.DataFrame): The data to use for inference.
        device (str): The device to perform computations on.
        parameters (dict): The parameters dictionary.
        modelDir (str): The path to the directory containing the model to be used for inference.
        outputDir (str): The path to the directory where the output of the inference session will be stored.
    """
    # Defining our model here according to parameters mentioned in the configuration file
    print("Current model type : ", parameters["model"]["type"])
    print("Number of dims     : ", parameters["model"]["dimension"])
    if "num_channels" in parameters["model"]:
        print("Number of channels : ", parameters["model"]["num_channels"])
    print("Number of classes  : ", len(parameters["model"]["class_list"]))
    parameters["testing_data"] = inferenceDataFromPickle

    # ensure outputs are saved properly
    parameters["save_output"] = True

    assert (
        parameters["model"]["type"].lower() == "torch"
        or parameters["model"]["type"].lower() == "openvino"
    ), f"The model type is not recognized: {parameters['model']['type']}"

    (model, _, _, _, _, parameters) = create_pytorch_objects(parameters, device=device)

    # Loading the weights into the model
    main_dict = None
    if parameters["model"]["type"].lower() == "torch":
        # Loading the weights into the model
        if os.path.isdir(modelDir):
            files_to_check = [
                os.path.join(
                    modelDir,
                    str(parameters["model"]["architecture"]) + BEST_MODEL_PATH_END,
                ),
                os.path.join(
                    modelDir,
                    str(parameters["model"]["architecture"]) + LATEST_MODEL_PATH_END,
                ),
            ]

            file_to_load = None
            for best_file in files_to_check:
                if os.path.isfile(best_file):
                    file_to_load = best_file
                    break

            assert file_to_load != None, "The 'best_file' was not found"

        main_dict = torch.load(file_to_load, map_location=parameters["device"])
        state_dict = main_dict["model_state_dict"]
        if parameters.get("differential_privacy"):
            # this is required for torch==1.11 and for DP inference
            new_state_dict = {}
            for key, val in state_dict.items():
                new_key = key.replace("_module.", "")
                new_state_dict[new_key] = val  # remove `module.`
            state_dict = new_state_dict

        model.load_state_dict(state_dict)
        parameters["previous_parameters"] = main_dict.get("parameters", None)
        model.eval()
    elif parameters["model"]["type"].lower() == "openvino":
        # Loading the executable OpenVINO model
        if os.path.isdir(modelDir):
            xml_to_check = os.path.join(
                modelDir, str(parameters["model"]["architecture"]) + "_best.xml"
            )
            bin_to_check = os.path.join(
                modelDir, str(parameters["model"]["architecture"]) + "_best.bin"
            )
            if not os.path.isfile(xml_to_check):
                raise ValueError(
                    "The specified model IR was not found: {0}.".format(xml_to_check)
                )
            if not os.path.isfile(bin_to_check):
                raise ValueError(
                    "The model specified model weights was not found: {0}.".format(
                        bin_to_check
                    )
                )
            model, input_blob, output_blob = load_ov_model(xml_to_check, device.upper())
            parameters["model"]["IO"] = [input_blob, output_blob]

    if not (os.environ.get("HOSTNAME") is None):
        print("\nHostname     :" + str(os.environ.get("HOSTNAME")), flush=True)

    # radiology inference
    if parameters["modality"] == "rad":
        if parameters["model"]["print_summary"]:
            print_model_summary(
                model,
                parameters["batch_size"],
                parameters["model"]["num_channels"],
                parameters["patch_size"],
                parameters["device"],
            )

        # Setting up the inference loader
        inference_loader = get_testing_loader(parameters)

        print("Data Samples: ", len(inference_loader.dataset), flush=True)

        average_epoch_valid_loss, average_epoch_valid_metric = validate_network(
            model, inference_loader, None, parameters, mode="inference"
        )
        print(average_epoch_valid_loss, average_epoch_valid_metric)
    elif parameters["modality"] in ["path", "histo"]:
        # set some defaults
        parameters["stride_size"] = parameters.get("stride_size", None)
        parameters["slide_level"] = parameters.get("slide_level", 0)
        parameters["mask_level"] = parameters.get(
            "mask_level", parameters["slide_level"]
        )
        parameters["blending_alpha"] = float(parameters.get("blending_alpha", 0.5))

        output_to_write = "SubjectID,x_coords,y_coords"
        if parameters["problem_type"] == "regression":
            output_to_write += ",output"
        elif parameters["problem_type"] == "classification":
            for n in range(parameters["model"]["num_classes"]):
                output_to_write += ",probability_" + str(n)
        output_to_write += "\n"

        # actual computation
        pbar = tqdm(inferenceDataFromPickle.iterrows())
        for _, row in pbar:
            subject_name = row[parameters["headers"]["subjectIDHeader"]]
            os_image = openslide.open_slide(
                row[parameters["headers"]["channelHeaders"]].values[0]
            )
            max_defined_slide_level = os_image.level_count - 1
            parameters["slide_level"] = min(
                parameters["slide_level"], max_defined_slide_level
            )
            parameters["slide_level"] = max(parameters["slide_level"], 0)
            level_width, level_height = os_image.level_dimensions[
                parameters["slide_level"]
            ]
            subject_dest_dir = os.path.join(outputDir, str(subject_name))
            Path(subject_dest_dir).mkdir(parents=True, exist_ok=True)

            try:
                count_map, probs_map = None, None
                count_map = np.zeros((level_height, level_width), dtype=np.uint8)
                # this can probably be made into a single multi-class probability map that functions for all workloads
                probs_map = np.zeros(
                    (parameters["model"]["num_classes"], level_height, level_width),
                    dtype=np.float16,
                )
            except Exception as e:
                print(
                    "Could not initialize count and probability maps for subject ID:",
                    subject_name,
                    "; Error:",
                    e,
                    flush=True,
                    file=sys.stderr,
                )

            patch_size = parameters["patch_size"]
            # patch size should be 2D for histology
            if len(patch_size) == 3:
                patch_size = patch_size[:-1]

            transform_requested = get_transforms_for_preprocessing(
                parameters, [], False, False
            )

            pbar.set_description(
                "Constructing loader for subject: " + str(subject_name)
            )

            patient_dataset_obj = InferTumorSegDataset(
                row[parameters["headers"]["channelHeaders"]].values[0],
                patch_size=patch_size,
                stride_size=parameters["stride_size"],
                selected_level=parameters["slide_level"],
                mask_level=parameters["mask_level"],
                transform=transform_requested,
            )

            dataloader = DataLoader(
                patient_dataset_obj,
                batch_size=1,
                shuffle=False,
                num_workers=parameters["q_num_workers"],
            )
            # update patch_size in case microns were requested
            patch_size = patient_dataset_obj.get_patch_size()

            if parameters["model"]["print_summary"]:
                print_model_summary(
                    model,
                    parameters["batch_size"],
                    parameters["model"]["num_channels"],
                    patch_size,
                    parameters["device"],
                )

            pbar.set_description(
                "Looping over patches for subject: " + str(subject_name)
            )

            for image_patches, (x_coords, y_coords) in dataloader:
                x_coords, y_coords = x_coords.numpy(), y_coords.numpy()
                if parameters["model"]["type"] == "torch":
                    if parameters["model"]["amp"]:
                        with autocast():
                            output = model(
                                image_patches.float().to(parameters["device"])
                            )
                    else:
                        output = model(image_patches.float().to(parameters["device"]))
                    output = output.detach().cpu().numpy()
                else:
                    output = model(
                        inputs={
                            parameters["model"]["IO"][0][0]: image_patches.float()
                            .cpu()
                            .numpy()
                        }
                    )[parameters["model"]["IO"][1][0]]

                for i in range(int(output.shape[0])):
                    if count_map is not None:
                        count_map[
                            y_coords[i] : y_coords[i] + patch_size[1],
                            x_coords[i] : x_coords[i] + patch_size[0],
                        ] += 1
                    output_to_write += (
                        str(subject_name)
                        + ","
                        + str(x_coords[i])
                        + ","
                        + str(y_coords[i])
                    )
                    for n in range(parameters["model"]["num_classes"]):
                        # This is a temporary fix for the segmentation problem for single class
                        if probs_map is not None:
                            probs_map[
                                n,
                                y_coords[i] : y_coords[i] + patch_size[1],
                                x_coords[i] : x_coords[i] + patch_size[0],
                            ] += output[i][n]
                        if parameters["problem_type"] != "segmentation":
                            output_to_write += "," + str(output[i][n])
                    output_to_write += "\n"

            # ensure probability map is scaled
            # reusing variables to save memory
            if probs_map is not None:
                probs_map = np.divide(probs_map, count_map)

                # Check if out_probs_map is greater than 1, print a warning
                if np.max(probs_map) > 1:
                    # Print a warning
                    print(
                        "Warning: Probability map is greater than 1, report the images to GaNDLF developers"
                    )

            if count_map is not None:
                count_map = np.array(count_map * 255, dtype=np.uint16)
                imsave(
                    os.path.join(
                        subject_dest_dir,
                        str(row[parameters["headers"]["subjectIDHeader"]])
                        + "_count.png",
                    ),
                    count_map,
                )

            if parameters["problem_type"] != "segmentation":
                output_file = os.path.join(subject_dest_dir, "predictions.csv")
                with open(output_file, "w") as f:
                    f.write(output_to_write)

            heatmaps = {}
            if probs_map is not None:
                try:
                    for n in range(parameters["model"]["num_classes"]):
                        heatmap_gray = np.array(probs_map[n, ...] * 255, dtype=np.uint8)
                        heatmaps[str(n) + "_jet"] = cv2.applyColorMap(
                            heatmap_gray, cv2.COLORMAP_JET
                        )
                        heatmaps[str(n) + "_turbo"] = cv2.applyColorMap(
                            heatmap_gray, cv2.COLORMAP_TURBO
                        )
                        heatmaps[str(n) + "_agni"] = applyCustomColorMap(heatmap_gray)

                        # save the segmentation maps
                        file_to_write = os.path.join(
                            subject_dest_dir, "seg_map_" + str(n) + ".png"
                        )

                        segmap = ((probs_map[n, ...] > 0.5).astype(np.uint8)) * 255

                        cv2.imwrite(file_to_write, segmap)

                    for key in heatmaps:
                        file_to_write = os.path.join(
                            subject_dest_dir, "probability_map" + key + ".png"
                        )
                        cv2.imwrite(file_to_write, heatmaps[key])

                        # this is needed because openslide returns an RGBA image
                        os_image_array = np.asarray(
                            os_image.read_region(
                                (0, 0),
                                parameters["slide_level"],
                                (level_width, level_height),
                            ).convert("RGB")
                        )
                        blended_image = cv2.addWeighted(
                            os_image_array,
                            parameters["blending_alpha"],
                            heatmaps[key],
                            1 - parameters["blending_alpha"],
                            0,
                        )

                        file_to_write = os.path.join(
                            subject_dest_dir, "probability_map_blended_" + key + ".png"
                        )
                        cv2.imwrite(file_to_write, blended_image)
                except Exception as ex:
                    print("Could not write heatmaps; error:", ex)



import ast
import traceback
from copy import deepcopy
from GANDLF.configuration.differential_privacy_config import DifferentialPrivacyConfig
from GANDLF.configuration.post_processing_config import PostProcessingConfig
from GANDLF.configuration.pre_processing_config import (
    HistogramMatchingConfig,
    PreProcessingConfig,
)
from GANDLF.data.post_process import postprocessing_after_reverse_one_hot_encoding
import numpy as np
import sys
from GANDLF.configuration.optimizer_config import OptimizerConfig, optimizer_dict_config
from GANDLF.configuration.patch_sampler_config import PatchSamplerConfig
from GANDLF.configuration.scheduler_config import (
    SchedulerConfig,
    schedulers_dict_config,
)
from GANDLF.configuration.utils import initialize_key, combine_models
from GANDLF.metrics import surface_distance_ids


def validate_loss_function(value) -> dict:
    if isinstance(value, dict):  # if this is a dict
        if len(value) > 0:  # only proceed if something is defined
            for key in value:  # iterate through all keys
                if key == "mse":
                    if (value[key] is None) or not ("reduction" in value[key]):
                        value[key] = {}
                        value[key]["reduction"] = "mean"
                else:
                    # use simple string for other functions - can be extended with parameters, if needed
                    value = key
    else:
        if value == "focal":
            value = {"focal": {}}
            value["focal"]["gamma"] = 2.0
            value["focal"]["size_average"] = True
        elif value == "mse":
            value = {"mse": {}}
            value["mse"]["reduction"] = "mean"

    return value


def validate_metrics(value) -> dict:
    if not isinstance(value, dict):
        temp_dict = {}
    else:
        temp_dict = value

    # initialize metrics dict
    for metric in value:
        # assigning a new variable because some metrics can be dicts, and we want to get the first key
        comparison_string = metric
        if isinstance(metric, dict):
            comparison_string = list(metric.keys())[0]
        # these metrics always need to be dicts
        if comparison_string in [
            "accuracy",
            "f1",
            "precision",
            "recall",
            "specificity",
            "iou",
        ]:
            if not isinstance(metric, dict):
                temp_dict[metric] = {}
            else:
                temp_dict[comparison_string] = metric
        elif not isinstance(metric, dict):
            temp_dict[metric] = None

        # special case for accuracy, precision, recall, and specificity; which could be dicts
        ## need to find a better way to do this
        if any(
            _ in comparison_string
            for _ in ["precision", "recall", "specificity", "accuracy", "f1"]
        ):
            if comparison_string != "classification_accuracy":
                temp_dict[comparison_string] = initialize_key(
                    temp_dict[comparison_string], "average", "weighted"
                )
                temp_dict[comparison_string] = initialize_key(
                    temp_dict[comparison_string], "multi_class", True
                )
                temp_dict[comparison_string] = initialize_key(
                    temp_dict[comparison_string], "mdmc_average", "samplewise"
                )
                temp_dict[comparison_string] = initialize_key(
                    temp_dict[comparison_string], "threshold", 0.5
                )
                if comparison_string == "accuracy":
                    temp_dict[comparison_string] = initialize_key(
                        temp_dict[comparison_string], "subset_accuracy", False
                    )
        elif "iou" in comparison_string:
            temp_dict["iou"] = initialize_key(
                temp_dict["iou"], "reduction", "elementwise_mean"
            )
            temp_dict["iou"] = initialize_key(temp_dict["iou"], "threshold", 0.5)
        elif comparison_string in surface_distance_ids:
            temp_dict[comparison_string] = initialize_key(
                temp_dict[comparison_string], "connectivity", 1
            )
            temp_dict[comparison_string] = initialize_key(
                temp_dict[comparison_string], "threshold", None
            )

    value = temp_dict
    return value


def validate_class_list(value):
    if isinstance(value, str):
        if ("||" in value) or ("&&" in value):
            # special case for multi-class computation - this needs to be handled during one-hot encoding mask construction
            print(
                "WARNING: This is a special case for multi-class computation, where different labels are processed together, `reverse_one_hot` will need mapping information to work correctly"
            )
            temp_class_list = value
            # we don't need the brackets
            temp_class_list = temp_class_list.replace("[", "")
            temp_class_list = temp_class_list.replace("]", "")
            value = temp_class_list.split(",")
        else:
            try:
                value = ast.literal_eval(value)
                return value
            except Exception as e:
                assert (
                    False
                ), f"Could not evaluate the `class_list` in `model`, Exception: {str(e)}, {traceback.format_exc()}"
                # logging.error(
                #     f"Could not evaluate the `class_list` in `model`, Exception: {str(e)}, {traceback.format_exc()}"
                # )
    return value


def validate_patch_size(patch_size, dimension) -> list:
    if isinstance(patch_size, int) or isinstance(patch_size, float):
        patch_size = [patch_size]
    if len(patch_size) == 1 and dimension is not None:
        actual_patch_size = []
        for _ in range(dimension):
            actual_patch_size.append(patch_size[0])
        patch_size = actual_patch_size
    if len(patch_size) == 2:  # 2d check
        # ensuring same size during torchio processing
        patch_size.append(1)
        if dimension is None:
            dimension = 2
    elif len(patch_size) == 3:  # 2d check
        if dimension is None:
            dimension = 3
    return [patch_size, dimension]


def validate_norm_type(norm_type, architecture):
    if norm_type is None or norm_type.lower() == "none":
        if not ("vgg" in architecture):
            raise ValueError(
                "Normalization type cannot be 'None' for non-VGG architectures"
            )
    return norm_type


def validate_parallel_compute_command(value):
    parallel_compute_command = value
    parallel_compute_command = parallel_compute_command.replace("'", "")
    parallel_compute_command = parallel_compute_command.replace('"', "")
    value = parallel_compute_command
    return value


def validate_scheduler(value, learning_rate, num_epochs):
    if isinstance(value, str):
        value = SchedulerConfig(type=value)
    # Find the scheduler_config class based on the type
    combine_scheduler_class = schedulers_dict_config[value.type]
    # Combine it with the SchedulerConfig class
    schedulerConfigCombine = combine_models(SchedulerConfig, combine_scheduler_class)
    combineScheduler = schedulerConfigCombine(**value.model_dump())
    value = SchedulerConfig(**combineScheduler.model_dump())

    if value.type == "triangular":
        if value.max_lr is None:
            value.max_lr = learning_rate

    if value.type in [
        "reduce_on_plateau",
        "reduce-on-plateau",
        "plateau",
        "exp_range",
        "triangular",
    ]:
        if value.min_lr is None:
            value.min_lr = learning_rate * 0.001

    if value.type in ["warmupcosineschedule", "wcs"]:
        value.warmup_steps = num_epochs * 0.1

    if hasattr(value, "step_size") and value.step_size is None:
        value.step_size = learning_rate / 5.0

    return value


def validate_optimizer(value):
    if isinstance(value, str):
        value = OptimizerConfig(type=value)

    combine_optimizer_class = optimizer_dict_config[value.type]
    # Combine it with the OptimizerConfig class
    optimizerConfigCombine = combine_models(OptimizerConfig, combine_optimizer_class)
    combineOptimizer = optimizerConfigCombine(**value.model_dump())
    value = OptimizerConfig(**combineOptimizer.model_dump())

    return value


def validate_data_preprocessing(value) -> dict:
    if not (value is None):
        # perform this only when pre-processing is defined
        if len(value) > 0:
            thresholdOrClip = False
            # this can be extended, as required
            thresholdOrClipDict = ["threshold", "clip", "clamp"]

            resize_requested = False
            temp_dict = deepcopy(value)
            for key in value:
                if key in ["resize", "resize_image", "resize_images", "resize_patch"]:
                    resize_requested = True

                if key in ["resample_min", "resample_minimum"]:
                    if "resolution" in value[key]:
                        resize_requested = True
                        resolution_temp = np.array(value[key]["resolution"])
                        if resolution_temp.size == 1:
                            temp_dict[key]["resolution"] = np.array(
                                [resolution_temp, resolution_temp]
                            ).tolist()
                    else:
                        temp_dict.pop(key)

            value = temp_dict

            if resize_requested and "resample" in value:
                for key in ["resize", "resize_image", "resize_images", "resize_patch"]:
                    if key in value:
                        value.pop(key)

                print(
                    "WARNING: Different 'resize' operations are ignored as 'resample' is defined under 'data_processing'",
                    file=sys.stderr,
                )

            # iterate through all keys
            for key in value:  # iterate through all keys
                if key in thresholdOrClipDict:
                    # we only allow one of threshold or clip to occur and not both
                    assert not (
                        thresholdOrClip
                    ), "Use only `threshold` or `clip`, not both"
                    thresholdOrClip = True
                    # initialize if nothing is present
                    if not (isinstance(value[key], dict)):
                        value[key] = {}

                    # if one of the required parameters is not present, initialize with lowest/highest possible values
                    # this ensures the absence of a field doesn't affect processing
                    # for threshold or clip, ensure min and max are defined
                    if not "min" in value[key]:
                        value[key]["min"] = sys.float_info.min
                    if not "max" in value[key]:
                        value[key]["max"] = sys.float_info.max

            key = "histogram_matching"
            if key in value:
                if value["histogram_matching"] is not False:
                    if not (isinstance(value["histogram_matching"], dict)):
                        value["histogram_matching"] = HistogramMatchingConfig()

            key = "histogram_equalization"
            if key in value:
                if value[key] is not False:
                    # if histogram equalization is enabled, call histogram_matching
                    value["histogram_matching"] = HistogramMatchingConfig()
            key = "adaptive_histogram_equalization"
            if key in value:
                if value[key] is not False:
                    # if histogram equalization is enabled, call histogram_matching
                    value["histogram_matching"] = HistogramMatchingConfig(
                        target="adaptive"
                    )

    pre_processing = PreProcessingConfig(**value)
    return pre_processing.model_dump(include={field for field in value.keys()})


def validate_data_postprocessing_after_reverse_one_hot_encoding(
    value, data_postprocessing
) -> list:
    temp_dict = deepcopy(value)
    for key in temp_dict:
        if key in postprocessing_after_reverse_one_hot_encoding:
            value[key] = data_postprocessing[key]
            data_postprocessing.pop(key)
    return [value, data_postprocessing]


def validate_patch_sampler(value):
    if isinstance(value, str):
        value = PatchSamplerConfig(type=value.lower())
    return value


def validate_data_augmentation(value, patch_size) -> dict:
    value["default_probability"] = value.get("default_probability", 0.5)
    if not (value is None):
        if len(value) > 0:  # only when augmentations are defined
            # special case for random swapping and elastic transformations - which takes a patch size for computation
            for key in ["swap", "elastic"]:
                if key in value:
                    value[key] = initialize_key(
                        value[key],
                        "patch_size",
                        np.round(np.array(patch_size) / 10).astype("int").tolist(),
                    )

            # special case for swap default initialization
            if "swap" in value:
                value["swap"] = initialize_key(value["swap"], "num_iterations", 100)

            # special case for affine default initialization
            if "affine" in value:
                value["affine"] = initialize_key(value["affine"], "scales", 0.1)
                value["affine"] = initialize_key(value["affine"], "degrees", 15)
                value["affine"] = initialize_key(value["affine"], "translation", 2)

            if "motion" in value:
                value["motion"] = initialize_key(value["motion"], "num_transforms", 2)
                value["motion"] = initialize_key(value["motion"], "degrees", 15)
                value["motion"] = initialize_key(value["motion"], "translation", 2)
                value["motion"] = initialize_key(
                    value["motion"], "interpolation", "linear"
                )

            # special case for random blur/noise - which takes a std-dev range
            for std_aug in ["blur", "noise_var"]:
                if std_aug in value:
                    value[std_aug] = initialize_key(value[std_aug], "std", None)
            for std_aug in ["noise"]:
                if std_aug in value:
                    value[std_aug] = initialize_key(value[std_aug], "std", [0, 1])

            # special case for random noise - which takes a mean range
            for mean_aug in ["noise", "noise_var"]:
                if mean_aug in value:
                    value[mean_aug] = initialize_key(value[mean_aug], "mean", 0)

            # special case for augmentations that need axis defined
            for axis_aug in ["flip", "anisotropic", "rotate_90", "rotate_180"]:
                if axis_aug in value:
                    value[axis_aug] = initialize_key(value[axis_aug], "axis", [0, 1, 2])

            # special case for colorjitter
            if "colorjitter" in value:
                value = initialize_key(value, "colorjitter", {})
                for key in ["brightness", "contrast", "saturation"]:
                    value["colorjitter"] = initialize_key(
                        value["colorjitter"], key, [0, 1]
                    )
                value["colorjitter"] = initialize_key(
                    value["colorjitter"], "hue", [-0.5, 0.5]
                )

            # Added HED augmentation in gandlf
            hed_augmentation_types = [
                "hed_transform",
                # "hed_transform_light",
                # "hed_transform_heavy",
            ]
            for augmentation_type in hed_augmentation_types:
                if augmentation_type in value:
                    value = initialize_key(value, "hed_transform", {})
                    ranges = [
                        "haematoxylin_bias_range",
                        "eosin_bias_range",
                        "dab_bias_range",
                        "haematoxylin_sigma_range",
                        "eosin_sigma_range",
                        "dab_sigma_range",
                    ]

                    default_range = (
                        [-0.1, 0.1]
                        if augmentation_type == "hed_transform"
                        else (
                            [-0.03, 0.03]
                            if augmentation_type == "hed_transform_light"
                            else [-0.95, 0.95]
                        )
                    )

                    for key in ranges:
                        value["hed_transform"] = initialize_key(
                            value["hed_transform"], key, default_range
                        )

                    value["hed_transform"] = initialize_key(
                        value["hed_transform"], "cutoff_range", [0, 1]
                    )

            # special case for anisotropic
            if "anisotropic" in value:
                if not ("downsampling" in value["anisotropic"]):
                    default_downsampling = 1.5
                else:
                    default_downsampling = value["anisotropic"]["downsampling"]

                initialize_downsampling = False
                if isinstance(default_downsampling, list):
                    if len(default_downsampling) != 2:
                        initialize_downsampling = True
                        print(
                            "WARNING: 'anisotropic' augmentation needs to be either a single number of a list of 2 numbers: https://torchio.readthedocs.io/transforms/augmentation.html?highlight=randomswap#torchio.transforms.RandomAnisotropy.",
                            file=sys.stderr,
                        )
                        default_downsampling = default_downsampling[0]  # only
                else:
                    initialize_downsampling = True

                if initialize_downsampling:
                    if default_downsampling < 1:
                        print(
                            "WARNING: 'anisotropic' augmentation needs the 'downsampling' parameter to be greater than 1, defaulting to 1.5.",
                            file=sys.stderr,
                        )
                        # default
                    value["anisotropic"]["downsampling"] = 1.5

            for key in value:
                if key != "default_probability":
                    value[key] = initialize_key(
                        value[key], "probability", value["default_probability"]
                    )
    return value


def validate_postprocessing(value):
    post_processing = PostProcessingConfig(**value)
    return post_processing.model_dump(include={field for field in value.keys()})


def validate_differential_privacy(value, batch_size):
    if value is None:
        return value
    if not isinstance(value, dict):
        print(
            "WARNING: Non dictionary value for the key: 'differential_privacy' was used, replacing with default valued dictionary."
        )
        value = DifferentialPrivacyConfig(physical_batch_size=batch_size).model_dump()
    # these are some defaults

    if value["physical_batch_size"] > batch_size:
        print(
            f"WARNING: The physical batch size {value['physical_batch_size']} is greater"
            f"than the batch size {batch_size}, setting the physical batch size to the batch size."
        )
    value["physical_batch_size"] = batch_size

    # these keys need to be parsed as floats, not strings
    for key in ["noise_multiplier", "max_grad_norm", "delta", "epsilon"]:
        if key in value:
            value[key] = float(value[key])

    return DifferentialPrivacyConfig(**value)



from .modelBase import ModelBase
import monai.networks.nets.dynunet as dynunet


def get_kernels_strides(sizes, spacings):
    """
    More info: https://github.com/Project-MONAI/tutorials/blob/main/modules/dynunet_pipeline/create_network.py#L19

    When refering this method for other tasks, please ensure that the patch size for each spatial dimension should
    be divisible by the product of all strides in the corresponding dimension.
    In addition, the minimal spatial size should have at least one dimension that has twice the size of
    the product of all strides. For patch sizes that cannot find suitable strides, an error will be raised.

    """
    input_size = sizes
    strides, kernels = [], []
    while True:
        spacing_ratio = [sp / min(spacings) for sp in spacings]
        stride = [
            2 if ratio <= 2 and size >= 8 else 1
            for (ratio, size) in zip(spacing_ratio, sizes)
        ]
        kernel = [3 if ratio <= 2 else 1 for ratio in spacing_ratio]
        if all(s == 1 for s in stride):
            break
        for idx, (i, j) in enumerate(zip(sizes, stride)):
            assert (
                i % j == 0
            ), f"Patch size is not supported, please try to modify the size {input_size[idx]} in the spatial dimension {idx}."
        sizes = [i / j for i, j in zip(sizes, stride)]
        spacings = [i * j for i, j in zip(spacings, stride)]
        kernels.append(kernel)
        strides.append(stride)

    strides.insert(0, len(spacings) * [1])
    kernels.append(len(spacings) * [3])
    return kernels, strides


class dynunet_wrapper(ModelBase):
    """
    More info: https://docs.monai.io/en/stable/networks.html#dynunet

    Args:
        spatial_dims (int): number of spatial dimensions.
        in_channels (int): number of input channels.
        out_channels (int): number of output channels.
        kernel_size (Sequence[Union[Sequence[int], int]]): convolution kernel size.
        strides (Sequence[Union[Sequence[int], int]]): convolution strides for each blocks.
        upsample_kernel_size (Sequence[Union[Sequence[int], int]]): convolution kernel size for transposed convolution layers. The values should equal to strides[1:].
        filters (Optional[Sequence[int]]): number of output channels for each blocks.  Defaults to None.
        dropout (Union[Tuple, str, float, None]): dropout ratio. Defaults to no dropout.
        norm_name (Union[Tuple, str]): feature normalization type and arguments. Defaults to INSTANCE.
        act_name (Union[Tuple, str]): activation layer type and arguments. Defaults to leakyrelu.
        deep_supervision (bool): whether to add deep supervision head before output. Defaults to False.
        deep_supr_num (int): number of feature maps that will output during deep supervision head. The value should be larger than 0 and less than the number of up sample layers. Defaults to 1.
        res_block (bool): whether to use residual connection based convolution blocks during the network. Defaults to False.
        trans_bias (bool): whether to set the bias parameter in transposed convolution layers. Defaults to False.
    """

    def __init__(self, parameters: dict):
        super(dynunet_wrapper, self).__init__(parameters)

        patch_size = parameters.get("patch_size", None)
        spacing = parameters.get(
            "spacing_for_internal_computations",
            [1.0 for i in range(parameters["model"]["dimension"])],
        )
        parameters["model"]["kernel_size"] = parameters["model"].get(
            "kernel_size", None
        )
        parameters["model"]["strides"] = parameters["model"].get("strides", None)
        if (parameters["model"]["kernel_size"] is None) or (
            parameters["model"]["strides"] is None
        ):
            kernel_size, strides = get_kernels_strides(patch_size, spacing)
            parameters["model"]["kernel_size"] = kernel_size
            parameters["model"]["strides"] = strides

        parameters["model"]["filters"] = parameters["model"].get("filters", None)
        parameters["model"]["act_name"] = parameters["model"].get(
            "act_name", ("leakyrelu", {"inplace": True, "negative_slope": 0.01})
        )
        parameters["model"]["deep_supervision"] = parameters["model"].get(
            "deep_supervision", False
        )
        parameters["model"]["deep_supr_num"] = parameters["model"].get(
            "deep_supr_num", 1
        )
        parameters["model"]["res_block"] = parameters["model"].get("res_block", True)
        parameters["model"]["trans_bias"] = parameters["model"].get("trans_bias", False)
        parameters["model"]["dropout"] = parameters["model"].get("dropout", None)

        if not ("norm_type" in parameters["model"]):
            self.norm_type = "INSTANCE"

        self.model = dynunet.DynUNet(
            spatial_dims=self.n_dimensions,
            in_channels=self.n_channels,
            out_channels=self.n_classes,
            kernel_size=parameters["model"]["kernel_size"],
            strides=parameters["model"]["strides"],
            upsample_kernel_size=parameters["model"]["strides"][1:],
            filters=parameters["model"][
                "filters"
            ],  # number of output channels for each blocks
            dropout=parameters["model"][
                "dropout"
            ],  # dropout ratio. Defaults to no dropout
            norm_name=self.norm_type,
            act_name=parameters["model"]["act_name"],
            deep_supervision=parameters["model"]["deep_supervision"],
            deep_supr_num=parameters["model"][
                "deep_supr_num"
            ],  # number of feature maps that will output during deep supervision head.
            res_block=parameters["model"]["res_block"],
            trans_bias=parameters["model"]["trans_bias"],
        )

    def forward(self, x):
        return self.model.forward(x)



from typing import Optional, Tuple
from pandas.util import hash_pandas_object
import torch
from torch.utils.data import DataLoader
from abc import ABC, abstractmethod
from dataclasses import dataclass
from GANDLF.models import get_model
from GANDLF.schedulers import get_scheduler
from GANDLF.optimizers import get_optimizer
from GANDLF.data import get_train_loader, get_validation_loader, ImagesFromDataFrame
from GANDLF.utils import (
    populate_header_in_parameters,
    populate_channel_keys_in_params,
    parseTrainingCSV,
    send_model_to_device,
    get_class_imbalance_weights,
)
from GANDLF.utils.write_parse import get_dataframe
from torchio import SubjectsDataset, Queue
from typing import Union


@dataclass
class AbstractSubsetDataParser(ABC):
    """
    Interface for subset data parsers, needed to separate the dataset creation
    from construction of the dataloaders.
    """

    subset_csv_path: str
    parameters_dict: dict

    @abstractmethod
    def create_subset_dataset(self) -> Union[SubjectsDataset, Queue]:
        """
        Method to create the subset dataset based on the subset CSV file
        and the parameters dict.

        Returns:
            Union[SubjectsDataset, Queue]: The subset dataset.
        """
        pass

    def get_params_extended_with_subset_data(self) -> dict:
        """
        Trick to get around the fact that parameters dict need to be modified
        during this parsing procedure. This method should be called after
        create_subset_dataset(), as this method will populate the parameters
        dict with the headers from the subset data.
        """
        return self.parameters_dict


class TrainingSubsetDataParser(AbstractSubsetDataParser):
    def create_subset_dataset(self) -> Union[SubjectsDataset, Queue]:
        (
            self.parameters_dict["training_data"],
            headers_to_populate_train,
        ) = parseTrainingCSV(self.subset_csv_path, train=True)

        self.parameters_dict = populate_header_in_parameters(
            self.parameters_dict, headers_to_populate_train
        )

        (
            self.parameters_dict["penalty_weights"],
            self.parameters_dict["sampling_weights"],
            self.parameters_dict["class_weights"],
        ) = get_class_imbalance_weights(
            self.parameters_dict["training_data"], self.parameters_dict
        )

        print("Penalty weights : ", self.parameters_dict["penalty_weights"])
        print("Sampling weights: ", self.parameters_dict["sampling_weights"])
        print("Class weights   : ", self.parameters_dict["class_weights"])

        return ImagesFromDataFrame(
            get_dataframe(self.parameters_dict["training_data"]),
            self.parameters_dict,
            train=True,
            loader_type="train",
        )


class ValidationSubsetDataParser(AbstractSubsetDataParser):
    def create_subset_dataset(self) -> Union[SubjectsDataset, Queue]:
        (self.parameters_dict["validation_data"], _) = parseTrainingCSV(
            self.subset_csv_path, train=False
        )
        validation_dataset = ImagesFromDataFrame(
            get_dataframe(self.parameters_dict["validation_data"]),
            self.parameters_dict,
            train=False,
            loader_type="validation",
        )
        self.parameters_dict = populate_channel_keys_in_params(
            validation_dataset, self.parameters_dict
        )
        return validation_dataset


class TestSubsetDataParser(AbstractSubsetDataParser):
    def create_subset_dataset(self) -> Union[SubjectsDataset, Queue]:
        testing_dataset = ImagesFromDataFrame(
            get_dataframe(self.subset_csv_path),
            self.parameters_dict,
            train=False,
            loader_type="testing",
        )
        if not ("channel_keys" in self.parameters_dict):
            self.parameters_dict = populate_channel_keys_in_params(
                testing_dataset, self.parameters_dict
            )
        return testing_dataset


class InferenceSubsetDataParserRadiology(TestSubsetDataParser):
    """Simple wrapper for name coherency, functionally this is the same as TestSubsetDataParser"""

    pass


def create_pytorch_objects(
    parameters: dict,
    train_csv: Optional[str] = None,
    val_csv: Optional[str] = None,
    device: Optional[str] = "cpu",
) -> Tuple[
    torch.nn.Module,
    torch.optim.Optimizer,
    DataLoader,
    DataLoader,
    torch.optim.lr_scheduler.LRScheduler,
    dict,
]:
    """
    This function creates the PyTorch objects needed for training and validation.

    Args:
        parameters (dict): The parameters for the model and training.
        train_csv (Optional[str], optional): The path to the training CSV file. Defaults to None.
        val_csv (Optional[str], optional): The path to the validation CSV file. Defaults to None.
        device (Optional[str], optional): The device to use for training. Defaults to "cpu".

    Returns:
        Tuple[ torch.nn.Module, torch.optim.Optimizer, DataLoader, DataLoader, torch.optim.lr_scheduler.LRScheduler, dict, ]: The model, optimizer, train loader, validation loader, scheduler, and parameters.
    """
    # initialize train and val loaders
    train_loader, val_loader = None, None
    headers_to_populate_train, headers_to_populate_val = None, None

    if train_csv is not None:
        # populate the data frames
        parameters["training_data"], headers_to_populate_train = parseTrainingCSV(
            train_csv, train=True
        )
        parameters = populate_header_in_parameters(
            parameters, headers_to_populate_train
        )

        # Calculate the weights here
        (
            parameters["penalty_weights"],
            parameters["sampling_weights"],
            parameters["class_weights"],
        ) = get_class_imbalance_weights(parameters["training_data"], parameters)

        print("Penalty weights : ", parameters["penalty_weights"])
        print("Sampling weights: ", parameters["sampling_weights"])
        print("Class weights   : ", parameters["class_weights"])

        # get the train loader
        train_loader = get_train_loader(parameters)
        parameters["training_samples_size"] = len(train_loader)
        # get the hash of the training data for reproducibility
        parameters["training_data_hash"] = hash_pandas_object(
            parameters["training_data"]
        ).sum()

    if val_csv is not None:
        parameters["validation_data"], headers_to_populate_val = parseTrainingCSV(
            val_csv, train=False
        )
        if headers_to_populate_train is None:
            parameters = populate_header_in_parameters(
                parameters, headers_to_populate_val
            )
        # get the validation loader
        val_loader = get_validation_loader(parameters)

    # get the model
    model = get_model(parameters)
    parameters["model_parameters"] = model.parameters()

    # get the optimizer
    optimizer = get_optimizer(parameters)
    parameters["optimizer_object"] = optimizer

    # send model to correct device
    (
        model,
        parameters["model"]["amp"],
        parameters["device"],
        parameters["device_id"],
    ) = send_model_to_device(
        model, amp=parameters["model"]["amp"], device=device, optimizer=optimizer
    )

    # only need to create scheduler if training
    if train_csv is not None:
        if not ("step_size" in parameters["scheduler"]):
            parameters["scheduler"]["step_size"] = (
                parameters["training_samples_size"] / parameters["learning_rate"]
            )

        scheduler = get_scheduler(parameters)

    else:
        scheduler = None

    # these keys contain generators, and are not needed beyond this point in params
    generator_keys_to_remove = ["optimizer_object", "model_parameters"]
    for key in generator_keys_to_remove:
        parameters.pop(key, None)

    return model, optimizer, train_loader, val_loader, scheduler, parameters



from .wrap_torch import (
    base_triangle,
    triangle_modified,
    cyclic_lr_base,
    cyclic_lr_exp_range,
    exp,
    step,
    reduce_on_plateau,
    cosineannealing,
)

from .wrap_monai import warmupcosineschedule

# defining dict for schedulers - key is the string and the value is the transform object
global_schedulers_dict = {
    "triangle": base_triangle,
    "triangle_modified": triangle_modified,
    "triangular": cyclic_lr_base,
    "exp_range": cyclic_lr_exp_range,
    "exp": exp,
    "exponential": exp,
    "step": step,
    "reduce_on_plateau": reduce_on_plateau,
    "reduce-on-plateau": reduce_on_plateau,
    "plateau": reduce_on_plateau,
    "reduceonplateau": reduce_on_plateau,
    "cosineannealing": cosineannealing,
    "warmupcosineschedule": warmupcosineschedule,
    "wcs": warmupcosineschedule,
}


def get_scheduler(params):
    """
    Function to get the scheduler definition.

    Args:
        params (dict): The parameters' dictionary.

    Returns:
        scheduler (object): The scheduler definition.
    """
    chosen_scheduler = params["scheduler"]["type"].lower()
    assert (
        chosen_scheduler in global_schedulers_dict
    ), f"Could not find the requested scheduler '{params['scheduler']['type']}'"
    return global_schedulers_dict[chosen_scheduler](params)
